package util

import (
	"encoding/json"
	"fmt"
	"reflect"
	"strconv"
)

func fieldToKeyValue(fieldValue reflect.Value, fieldName string) (key, value string, err error) {
	switch fieldValue.Kind() {
	case reflect.Int8, reflect.Int, reflect.Int16, reflect.Int32, reflect.Int64:
		return fieldName, fmt.Sprintf("%v", fieldValue.Int()), nil
	case reflect.Uint8, reflect.Uint, reflect.Uint16, reflect.Uint32, reflect.Uint64:
		return fieldName, fmt.Sprintf("%v", fieldValue.Uint()), nil
	case reflect.Float32, reflect.Float64:
		return fieldName, fmt.Sprintf("%v", fieldValue.Float()), nil

	case reflect.String:
		return fieldName, fieldValue.String(), nil
	case reflect.Bool:
		return fieldName, strconv.FormatBool(fieldValue.Bool()), nil

	default:
		if str, err := json.Marshal(fieldValue.Interface()); err != nil {
			return "", "", err
		} else {
			return fieldName, string(str), nil
		}
	}
}

func Serialize(structObj interface{}) (map[string]string, error) {
	value := reflect.ValueOf(structObj)
	if value.Kind() == reflect.Ptr {
		value = value.Elem()
	}
	ret := make(map[string]string)
	// 确保它是一个结构体
	if value.Kind() == reflect.Struct {
		for i := 0; i < value.NumField(); i++ {
			// 获取字段的反射类型Value和Type
			fieldValue := value.Field(i)
			fieldType := value.Type().Field(i)
			fieldName := fieldType.Tag.Get("json")
			if fieldName == "-" {
				continue
			}
			if k, v, err := fieldToKeyValue(fieldValue, fieldName); err == nil {
				ret[k] = v
			} else {
				return nil, err
			}
		}
	} else {
		return ret, fmt.Errorf("input should be struct or struct pointer")
	}
	return ret, nil
}

func keyValueToField(key, value string, fieldValue *reflect.Value) error {
	switch fieldValue.Kind() {
	case reflect.Int8, reflect.Int, reflect.Int16, reflect.Int32, reflect.Int64:
		if v, err := strconv.ParseInt(value, 10, 0); err == nil {
			fieldValue.SetInt(v)
		} else {
			return err
		}
	case reflect.Uint8, reflect.Uint, reflect.Uint16, reflect.Uint32, reflect.Uint64:
		if v, err := strconv.ParseUint(value, 10, 0); err == nil {
			fieldValue.SetUint(v)
		} else {
			return err
		}
	case reflect.Float32, reflect.Float64:
		if v, err := strconv.ParseFloat(value, 0); err == nil {
			fieldValue.SetFloat(v)
		} else {
			return err
		}
	case reflect.String:
		fieldValue.SetString(value)
	case reflect.Bool:
		if v, err := strconv.ParseBool(value); err == nil {
			fieldValue.SetBool(v)
		} else {
			return err
		}
	default:
		nv := reflect.New(fieldValue.Type())
		if err := json.Unmarshal([]byte(value), nv.Interface()); err == nil {
			if fieldValue.Kind() == nv.Elem().Kind() {
				fieldValue.Set(nv.Elem())
			} else {
				return fmt.Errorf("field:%v should have same type.original type:%v, reflect type:%v", key, fieldValue.Kind().String(), nv.Elem().Kind().String())
			}
		} else {
			return err
		}
	}
	return nil
}

func Deserialize(kvs map[string]string, structObj interface{}) error {
	val := reflect.ValueOf(structObj).Elem()
	typ := val.Type()

	for i := 0; i < typ.NumField(); i++ {
		field := typ.Field(i)
		// 检查字段是否可设置（即导出）
		if !field.IsExported() {
			continue
		}
		fieldName := field.Tag.Get("json")
		if fieldName == "-" {
			continue
		}
		if value, ok := kvs[fieldName]; ok {
			// 获取字段的反射值
			fieldValue := val.FieldByName(field.Name)
			if err := keyValueToField(field.Tag.Get("json"), value, &fieldValue); err != nil {
				return err
			}
		} else {
			return fmt.Errorf("redis:%v not has field:%v ", typ.Name(), field.Tag.Get("json"))
		}
	}
	return nil
}
