package util

import (
	"encoding/base64"
	"encoding/json"
	"errors"
	"fmt"
	"reflect"
	"strings"
	"time"
)

// 将结构体转换为INSERT SQL语句
func struct2InsertSqlColumns(structTable interface{}) []string {
	val := reflect.ValueOf(structTable)
	if val.Kind() == reflect.Ptr {
		val = val.Elem()
	}
	typ := val.Type()
	// 构建字段名和值的列表
	var columns []string
	for i := 0; i < val.NumField(); i++ {
		// 忽略私有字段
		if typ.Field(i).PkgPath != "" {
			continue
		}

		// 查找json tag
		fieldName := typ.Field(i).Tag.Get("json")
		if fieldName == "" {
			// 如果没有指定tag，则使用字段名作为默认值
			fieldName = typ.Field(i).Name
		}
		// 添加到字段名和值列表中
		columns = append(columns, fieldName)
	}
	return columns
}

// 将结构体转换为INSERT SQL语句
func struct2InsertSqlValues[T any](structTable ...T) ([][]string, error) {
	var values [][]string
	for _, st := range structTable {
		val := reflect.ValueOf(st)
		if val.Kind() == reflect.Ptr {
			val = val.Elem()
		}
		typ := val.Type()
		var fieldValues []string
		for i := 0; i < val.NumField(); i++ {
			// 忽略私有字段
			if typ.Field(i).PkgPath != "" {
				continue
			}
			fieldValue := val.Field(i).Interface()
			switch val.Field(i).Interface().(type) {
			case int, int8, int16, int32, int64, uint, uint8, uint16, uint32, uint64:
				fieldValues = append(fieldValues, fmt.Sprintf("%v", fieldValue))
			case string:
				fieldValues = append(fieldValues, fmt.Sprintf("'%v'", fieldValue))
			case time.Time:
				fieldValues = append(fieldValues, fmt.Sprintf("'%v'", fieldValue.(time.Time).Format("2006-01-02 15:04:05")))
			default:
				if byteValue, err := json.Marshal(fieldValue); err == nil {
					b64Value := base64.StdEncoding.EncodeToString(byteValue)
					fieldValues = append(fieldValues, fmt.Sprintf("'%v'", b64Value))
				} else {
					return nil, err
				}
			}
		}
		values = append(values, fieldValues)
	}
	return values, nil
}

// 将结构体转换为INSERT SQL语句
func Struct2InsertSql[T any](tableName string, structTable ...T) (string, error) {
	if len(structTable) == 0 {
		return "", errors.New("struct table is empty")
	}
	var sql string
	columns := struct2InsertSqlColumns(structTable[0])
	groupValues, err := struct2InsertSqlValues(structTable...)
	if err != nil {
		return "", err
	}
	if len(groupValues) == 1 {
		sql = fmt.Sprintf("INSERT INTO %s (%s) VALUES (%s)", tableName, strings.Join(columns, ","), strings.Join(groupValues[0], ","))
	} else {
		content := ""
		for pos, values := range groupValues {
			if pos == len(groupValues)-1 {
				content += fmt.Sprintf("(%s) ", strings.Join(values, ","))
			} else {
				content += fmt.Sprintf("(%s), ", strings.Join(values, ","))
			}
		}
		sql = fmt.Sprintf("INSERT INTO %s (%s) VALUES %s", tableName, strings.Join(columns, ","), content)
	}
	return sql, nil
}

// map转sql语句，未做防sql注入
func MapToInsertSQL(tableName string, data map[string]interface{}) (string, error) {
	// 检查tableName是否为空
	if tableName == "" {
		return "", fmt.Errorf("tableName cannot be empty")
	}

	// 准备SQL语句的组成部分
	var columns []string
	var values []string

	// 遍历map以构建columns和values，并准备占位符
	for column, value := range data {
		columns = append(columns, fmt.Sprintf("`%s`", column))

		vType := reflect.TypeOf(value)
		switch vType.Kind() {
		case reflect.Uint8, reflect.Int8, reflect.Uint16, reflect.Int16, reflect.Uint32, reflect.Int32, reflect.Uint64, reflect.Int64, reflect.Uint, reflect.Int:
			values = append(values, fmt.Sprintf("%v", value))
		case reflect.String:
			values = append(values, fmt.Sprintf("'%v'", value))
		case reflect.Float32, reflect.Float64:
			intV := int64(value.(float64))
			values = append(values, fmt.Sprintf("%v", intV))
		default:
			return "", fmt.Errorf("column:%v value:%v can not cast type:%v", column, value, vType.String())
		}
	}
	sql := fmt.Sprintf("INSERT INTO %s (%s) VALUES (%s);", tableName, strings.Join(columns, ", "), strings.Join(values, ", "))
	return sql, nil
}
